<?php
//EditorHelper
/**
 * Copyright 2020 Kenneth Brill ken.brill@gmail.com
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 * documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
 * OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.<br>
 */
$e = new editorHelper();
if (!isset($_POST['run'])) {
	$smarty = new Sugar_Smarty();
	$smarty->assign('VERSION', $e->version);
	echo $smarty->fetch('custom/modules/Administration/editorHelper.tpl');
} else {
	switch ($_POST['mode']) {
		case 'Normal':
			$e->justRemove = false;
			$e->debugMode = false;
			$e->customFiles = false;
			break;
		case 'Debug':
			$e->debugMode = true;
			break;
		case 'Custom':
			$e->customFiles = true;
			break;
		case 'Clean':
			$e->justRemove = true;
			break;
	}
	$e->addCustomFieldsToClasses();
}

/**
 * Class editorHelper
 */
class editorHelper
{
	public $version = 1.15;
	private $utilityName = 'editorHelper';

	//Setting this to true will simply clean up all the files
	// and then stop.
	public $justRemove = false;

	//In Debug mode we write everything to separate debug files
	// in the core directories and not to existing files in Sugar
	// the files are named module_debug.php (ie modules/Accounts/Account_debug.php)
	public $debugMode = false;

	//This switch creates custom classes instead of updating the
	// core files. Its safer but then you have to use those custom classes
	// in your code to see the benefit.
	public $customFiles = false;

	private $modeArray = array('debug', 'custom');

	public function addCustomFieldsToClasses(): void
	{
		global $beanFiles, $beanList;
		$moduleNames = array_flip($beanList);
		//Remove mode is incompatible with debug and custom modes
		if ($this->justRemove) {
			$this->debugMode = false;
			$this->customFiles = false;
		}
		//if we are in debug mode then we dont really need to
		// create anything in the custom directory
		if ($this->debugMode) {
			$this->customFiles = false;
		}
		$this->cleanClassFiles();
		if (!$this->justRemove) {
			foreach ($beanFiles as $className => $fileName) {
				$addedFields = 0;
				$bean = BeanFactory::newBean($moduleNames[$className]);
				$customFileName = $this->getUpdatedFileName($fileName);
				if ($bean !== null) {
					//If we find a debug/custom file then we use that
					if (file_exists($customFileName)) {
						$fileName = $customFileName;
					}

					if (file_exists($fileName)) {
						$classFile = file($fileName);
						//Update the class name (if needed)
						$classFile = $this->updateClassName($classFile, $className);
						$insertionPoint = $this->findInsertionPoint($classFile, $className);
						if ($insertionPoint !== false) {
							if (isset($bean->field_defs)) {
								$field_defs = array_keys($bean->field_defs);
								if (!empty($field_defs)) {
									foreach ($field_defs as $fieldName) {
										if (substr($fieldName, -2) === '_c') {
											$addedFields++;
											$date = gmdate('Y-m-d H:i');
											$insertedLine = array("\tpublic \${$fieldName}; //Added by {$this->utilityName} on {$date} GMT" . PHP_EOL);
											array_splice($classFile, $insertionPoint, 0, $insertedLine);
										}
									}
									if ($addedFields > 0) {
										//After about a thousands tests, I found it was just best to delete the
										//file before writing to it.
										//In this section of the code we write to $customFileName instead of
										// $fileName even though we may hve been reading from $fileName
										if (file_exists($customFileName)) {
											if(is_readable($customFileName)) {
												$path_parts = pathinfo($customFileName);
												$newFileName = $path_parts['dirname'].'/'.$path_parts['filename'].'.bak';
												if(file_exists($newFileName)) {
													if(!unlink($newFileName)) {
														sugar_die("Could not delete the file {$newFileName}");
													}
												}
												if(!copy($customFileName, $newFileName)) {
													$this->output("Could not create backup file: {$newFileName}");
												}
											}
											if (is_writable($customFileName)) {
												if (!unlink($customFileName)) {
													sugar_die("Could not delete the file {$customFileName}");
												}
												sugar_file_put_contents($customFileName, $classFile);
												$this->output("Added {$addedFields} field definitions to {$customFileName}");
											} else {
												sugar_die("Cannot write to {$customFileName}");
											}
										} else {
											sugar_file_put_contents($customFileName, $classFile);
											$this->output("Added {$addedFields} field definitions to {$customFileName}");
										}
									}
								}
							}
						}
					} else {
						sugar_die("The file '{$fileName}' does not exist and it should.  Something is terribly wrong");
					}
				}
			}
		}
	}

	/**
	 * Removes the lines from all the class files that were added from this utility
	 */
	private function cleanClassFiles(): void
	{
		global $beanFiles;
		foreach ($beanFiles as $className => $fileName) {
			$foundLineToRemove = false;
			$cleanedFile = '';
			//Delete non-stock files
			foreach ($this->modeArray as $mode) {
				$customFileName = $this->getUpdatedFileName($fileName, $mode);
				if (file_exists($customFileName)) {
					if (!unlink($customFileName)) {
						sugar_die("Could not delete the file {$customFileName}");
					} else {
						$this->output("Deleted {$customFileName}");
					}
				}
			}

			//Load the core class file
			$classFile = file($fileName);
			foreach ($classFile as $line) {
				if (stripos($line, $this->utilityName) === false) {
					$cleanedFile .= $line;
				} else {
					$foundLineToRemove = true;
				}
			}
			if ($foundLineToRemove) {
				//After about a thousand tests, I found it was just best to delete the
				//file before writing to it.
				if (file_exists($fileName)) {
					if (is_writable($fileName)) {
						if (!unlink($fileName)) {
							sugar_die("Could not delete the file {$fileName}");
						}
						sugar_file_put_contents($fileName, $cleanedFile);
						$this->output("Cleaned {$fileName}.");
					} else {
						sugar_die("Cannot write to {$fileName}");
					}
				} else {
					sugar_file_put_contents($fileName, $cleanedFile);
					$this->output("Cleaned {$fileName}.");
				}
			}

		}
	}


	/**
	 * Logging strings to the web
	 * @param $string
	 */
	private function output($string): void
	{
		echo $string . '<br>' . PHP_EOL;
		ob_flush();
		flush();
	}

	/**
	 * Returns the index of the line BELOW the class declaration
	 * @param array $classFile
	 * @param string $className
	 * @return bool|int
	 */
	private function findInsertionPoint($classFile, $className)
	{
		foreach ($classFile as $index => $line) {
			if (stripos($line, "class {$className}") !== false) {
				if (strpos(trim($classFile[$index + 1]), '{') === 0) {
					//add one to the index to get below the curly brace
					++$index;
				}
				return $index + 1;
			}
			if (stripos($line, "class custom{$className}") !== false) {
				if (strpos(trim($classFile[$index + 1]), '{') === 0) {
					//add one to the index to get below the curly brace
					++$index;
				}
				return $index + 1;
			}
		}
		return false;
	}

	/**
	 * Changes the Class name in a Debug file so it doesnt screw up phpStorm
	 * @param array $classFile
	 * @param string $className
	 * @return array
	 */
	private function updateClassName($classFile, $className): array
	{
		foreach ($classFile as $index => $line) {
			if (stripos($line, "class {$className}") !== false) {
				$newLine = '';
				if ($this->debugMode) {
					$newLine = str_replace("class {$className}", "class {$className}_debug", $line);
				}
				if ($this->customFiles) {
					$newLine = "class custom{$className} extends {$className} {" . PHP_EOL;
				}
				if (!empty($newLine)) {
					$classFile[$index] = $newLine;
				}
			}
		}
		return $classFile;
	}

	/**
	 * @param string $fileName
	 * @param string $mode
	 * @return string
	 */
	private function getUpdatedFileName($fileName, $mode = ''): string
	{
		if(!empty($mode) && !in_array($mode, $this->modeArray, false)) {
			sugar_die("Invalid mode: {$mode}");
		}
		$path_parts = pathinfo($fileName);
		if ($this->debugMode || $mode === 'debug') {
			$returnVar = $path_parts['dirname'] . '/' . $path_parts['filename'] . '_debug.php';
		} else if ($this->customFiles || $mode === 'custom') {
			$returnVar = 'custom/' . $path_parts['dirname'] . '/' . 'custom' . $path_parts['basename'];
		} else {
			$returnVar = $fileName;
		}
		return $returnVar;
	}
}


